/*
Topdown Vorverarbeitendes Determinanten Schema                       Korrektheitstest Version

In  dieser  Version  verzichtet  der Laplace Algorithmus auf Verarbeitung einer Entwicklungs-
reihenfolge,  um  durch  Verzicht  auf  die  komplexen Routinen ein NACHVOLLZIEHBAR korrektes
Ergebnis zu liefern.
Es  werden 1.000.000 zufllige Matrizen in einer Gre zwischen 2x2 und 10x10 generiert, ihre
Determinanten nach Laplace Schema und TVDS berechnet und verglichen.
Anhand der Routinen sollte man erkennen, dass die Ergebnisse nicht geflscht sind!
                                                                             Viktor Engelmann
*/

#include<conio>       // gotoxy
#include<iostream>    // cout & cin
#include<iomanip>     // setw
#include<stdlib>      // random
#include<dos>         // Uhrzeit & Datum
#pragma hdrstop
#include"TVDStree.h"

int** matrix;
int dimension;
// ===================================================================================== swap
template<class T> inline __fastcall void swap(T& a, T& b)
   {
   T swap = a;
   a = b;
   b = swap;
   }
// ================================================================= Entwicklungs Reihenfolge
int* reihenfolge;
int* anzahl;
int Faktor;
int Bestimme_Reihenfolge()
   {
   for(int i = 0; i < dimension; i++)
      {
      anzahl[i] = 0;
      reihenfolge[i] = i;
      for(int j = 0; j < dimension; j++)
         if(matrix[i][j] == 0)
            anzahl[i]++;
      }
   int faktor = 1;
   for(int i = 1; i < dimension; i++)
      for(int j = 0; j < dimension - i; j++)
         if(anzahl[j] < anzahl[j+1])
            {
            swap(anzahl[j], anzahl[j+1]);
            swap(reihenfolge[j], reihenfolge[j+1]);
            faktor *= -1;
            }
   return(faktor);
   }
// ===================================================================================== TVDS
bool* n_tupel;
TVDS_tree* calculated = NULL;
int TVDS(int rekdepth = 0)
   {
   int ret = 0;

   int* leafcont = findleaf(calculated, n_tupel);
   if(leafcont != NULL)
      return(*(leafcont));

   TVDS_tree*leaf = lastleaf;
   if(rekdepth == dimension-2)                    // 2x2 Unterdeterminante explizit berechnen
      {
      int sa; for(sa = 0; sa < dimension; sa++) if(n_tupel[sa]) break;
      int sb; for(sb = sa+1; sb < dimension; sb++) if(n_tupel[sb]) break;
      int za = reihenfolge[dimension-2];
      int zb = reihenfolge[dimension-1];
      leaf->content = new(int);                          // berechnete Determinante speichern
      *(leaf->content) = matrix[za][sa] * matrix[zb][sb] - matrix[za][sb] * matrix[zb][sa];
      return(*(leaf->content));
      }

   int z = reihenfolge[rekdepth];                       // Zeile aus der Reihenfolge auslesen
   int faktor = 1;
   for(int i = 0; i < dimension; i++)                            // Schleife ber die Spalten
      if(n_tupel[i])
         {
         if(matrix[z][i] != 0)
            {
            n_tupel[i] = false;                                           // Spalte streichen
            ret += faktor * matrix[z][i] * TVDS(rekdepth+1);                     // Rekursion
            n_tupel[i] = true;                                            // Spalte freigeben
            }
         faktor *= -1;
         }

   leaf->content = new(int);                       // fertig bestimmte Determinante speichern
   *(leaf->content) = ret;
   return(ret);                                           // (Unter-)Determinante zurckgeben
   }

// ================================================================================== Laplace
bool* ospalten;
int Laplace(int rekdepth = 0)
   {
   int ret = 0;

   if(rekdepth == dimension-2)                    // 2x2 Unterdeterminante explizit berechnen
      {
      int sa; for(sa = 0; sa < dimension; sa++) if(ospalten[sa]) break;
      int sb; for(sb = sa+1; sb < dimension; sb++) if(ospalten[sb]) break;
      int za = dimension-2;
      int zb = dimension-1;
      return(matrix[za][sa] * matrix[zb][sb] - matrix[za][sb] * matrix[zb][sa]);
      }

   int faktor = -1;                                              // Schleife ber die Spalten
   for(int i = 0; i < dimension; i++)
      if(ospalten[i])
         {
         faktor *= -1;

         if(matrix[rekdepth][i] != 0)
            {
            ospalten[i] = false;                                          // Spalte streichen
            ret += faktor * Laplace(rekdepth+1) * matrix[rekdepth][i];           // Rekursion
            ospalten[i] = true;                                           // Spalte freigeben
            }
         }
   return(ret);                                           // (Unter-)Determinante zurckgeben
   }

// ===================================================================================== MAIN
void main()
   {
   long i, richtig = 0, falsch = 0;
   struct dosdate_t d;
   struct dostime_t t;
   _dos_getdate(&d);
   _dos_gettime(&t);

   cout << "\n  Korrektheitstest                    " << short(d.day) << "."<<short(d.month)
        << "." << short(d.year) << "    " << short(t.hour) << ":" << short(t.minute);
   cout << "\n  -------------------------------------------------------\n";
   cout << "  Determinanten berechnet nach TVDS und Laplace:        0\n";
   cout << "  bereinstimmende Ergebnisse:                          0\n";
   cout << "  Unterschiedliche Ergebnisse:                          0\n";

   #define maxdim 10
   ospalten = new(bool[maxdim]);
   n_tupel = new(bool[maxdim]);
   reihenfolge = new(int[maxdim]);
   anzahl = new(int[maxdim]);
   matrix = new(int*[maxdim]);
   randomize();
   for(int i = 0; i < maxdim; i++)
      {
      ospalten[i] = true;
      n_tupel[i] = true;
      matrix[i] = new(int[maxdim]);
      }

   for(i = 0; i < 1000000 && !kbhit();)
      {
      calculated = new(TVDS_tree);
      dimension = random(maxdim-2)+2;
      for(int k = 0; k < dimension; k++)
         for(int j = 0; j < dimension; j++)
            matrix[k][j] = random(10);

      int faktor = Bestimme_Reihenfolge();

      gotoxy(49,4);
      cout << setw(9) << ++i;
      if(faktor * TVDS() == Laplace())
         {gotoxy(49,5);
         cout << setw(9) << ++richtig;}
      else
         {gotoxy(49,6);
         cout << setw(9) << ++falsch;}
      delete(calculated);
      }

   delete(ospalten);
   delete(n_tupel);
   delete(reihenfolge);
   delete(anzahl);
   for(int i = 0; i < maxdim; i++)
      delete(matrix[i]);
   delete(matrix);
   cout.flush();
   cin.ignore(1);
   }
