/*
Topdown Vorverarbeitendes Determinanten Schema                           Zeitmessungs Version

Dies  ist  eine  abgespeckte  Version des Algorithmus, sie geht davon aus, dass  keine  0  in
der  Matrix vorkommt, denn sie ist sowieso dazu gedacht, die Laufzeiten bei voller Auslastung
zu messen.
Am  Anfang  des  Programms wird der Benutzer gefragt, wie lang die Algorithmen maximal laufen
drfen, d.h. sobald  ein  Algorithmus  einmal  fr eine  Berechnung lnger gebraucht hat, als
der  Benutzer  erlaubt  hatte, wird der Algorithmus fr grere Matrizen nicht mehr getestet.
Diese  letzte  Berechnung  wird  allerdings  ber  die  Zeitgrenze  hinaus  bis  zum  Schluss
durchgefhrt, erst  danach  wird  der  Algorithmus  deaktiviert ( => der Test  kann die Zeit-
grenze  sehr  weit berschreiten) weil erst nach Beendigung des Tests geprft wird, wie lange
der  Test  gedauert  hat (eine  permanente  Prfung  WHREND des Tests wrde zu viel Laufzeit
kosten)
                                                                             Viktor Engelmann
*/
#include<iostream>    // cout & cin
using namespace std;
#include<iomanip>     // setw
#include<stdlib>      // random
#include<sys\timeb>   // struct timeb & ftime
#pragma hdrstop
#include"TVDStree.h"

int** matrix;
int dimension;

// ===================================================================================== TVDS
bool* n_tupel;
TVDS_tree* calculated = NULL;

int TVDS(int rekdepth = 0)
   {
   int ret = 0;

   int* leafcont = findleaf(calculated, n_tupel);                               // TVDS-cut
   if(leafcont != NULL)
      return(*(leafcont));

   TVDS_tree* leaf = lastleaf;

   if(rekdepth == dimension-2)                    // 2x2 Unterdeterminante explizit berechnen
      {
      int sa; for(sa = 0; sa < dimension; sa++) if(n_tupel[sa]) break;
      int sb; for(sb = sa+1; sb < dimension; sb++) if(n_tupel[sb]) break;
      int za = dimension-2;
      int zb = dimension-1;
      leaf->content = new(int);
      *(leaf->content) = matrix[za][sa] * matrix[zb][sb] - matrix[za][sb] * matrix[zb][sa];
      return(*(leaf->content));
      }

   int faktor = 1;
   for(int i = 0; i < dimension; i++)                            // Schleife ber die Spalten
      {
      if(n_tupel[i])
         {
         n_tupel[i] = false;                                              // Spalte streichen
         ret += faktor * matrix[rekdepth][i] * TVDS(rekdepth+1);                 // Rekursion
         n_tupel[i] = true;                                               // Spalte freigeben
         }
      faktor *= -1;
      }

   leaf->content = new(int);
   *(leaf->content) = ret;
   return(ret);                                           // (Unter-)Determinante zurckgeben
   }

// ================================================================================== Laplace
bool* ospalten;
int Laplace(int rekdepth = 0)
   {
   int ret = 0;

   if(rekdepth == dimension-2)                    // 2x2 Unterdeterminante explizit berechnen
      {
      int sa; for(sa = 0; sa < dimension; sa++) if(ospalten[sa]) break;
      int sb; for(sb = sa+1; sb < dimension; sb++) if(ospalten[sb]) break;
      int za = dimension-2;
      int zb = dimension-1;
      return(matrix[za][sa] * matrix[zb][sb] - matrix[za][sb] * matrix[zb][sa]);
      }

   int faktor = 1;                                               // Schleife ber die Spalten
   for(int i = 0; i < dimension; i++)
      {
      if(ospalten[i])
         {
         ospalten[i] = false;                                             // Spalte streichen
         ret += faktor * Laplace(rekdepth+1) * matrix[rekdepth][i];              // Rekursion
         ospalten[i] = true;                                              // Spalte freigeben
         }
      faktor *= -1;
      }
   return(ret);                                           // (Unter-)Determinante zurckgeben
   }

// ===================================================================================== MAIN
struct timeb s,e;
int zg;
bool uselaplace = false;
bool usetvds = true;

void main()
   {
   cout << "Zeitgrenze (min) > ";
   cin >> zg;
   zg *= 60;

   cout << "\n\n      |    TVDS   |  Laplace\n------+-----------+----------\n";
   for(int i = 22; usetvds || uselaplace; i++)
      {
      dimension = i;
      calculated = new(TVDS_tree);
      ospalten = new(bool[dimension]);
      n_tupel = new(bool[dimension]);
      matrix = new(int*[dimension]);
      randomize();
      for(int i = 0; i < dimension; i++)
         {
         ospalten[i] = true;
         n_tupel[i] = true;
         matrix[i] = new(int[dimension]);
         for(int j = 0; j < dimension; j++)
            matrix[i][j] = random(9)+1;
         }

      cout << setw(2) << i << "x" << setw(2) << i << " | ";

      if(usetvds)
         {
         ftime(&s);
         TVDS();
         ftime(&e);
         cout << setw(9) << ((e.time*1000+e.millitm)-(s.time*1000+s.millitm)) << " | ";
         if(e.time - s.time > zg)
            usetvds = false;
         }
      else
         cout << "          | ";

      if(uselaplace)
         {
         ftime(&s);
         Laplace();
         ftime(&e);
         cout << setw(9) << ((e.time*1000+e.millitm)-(s.time*1000+s.millitm));
         if(e.time - s.time > zg)
            uselaplace = false;
         }
      cout << endl;

      delete(calculated);
      delete(ospalten);
      delete(n_tupel);
      for(int i = 0; i < dimension; i++)
         delete(matrix[i]);
      delete(matrix);
      }
   cin.ignore(2);
   }
